@@ -1,6 +1,6 @@
 /**
  * Object oriented C module to send ICMP and ICMPv6 `echo's.
- * Copyright (C) 2006-2011  Florian octo Forster <ff at octo.it>
+ * Copyright (C) 2006-2014  Florian octo Forster <ff at octo.it>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -74,13 +74,52 @@
 #include <sys/types.h>
 #endif
 
+#include <locale.h>
+#include <langinfo.h>
+
 #if USE_NCURSES
 # define NCURSES_OPAQUE 1
-# include <ncurses.h>
+/* http://newsgroups.derkeiler.com/Archive/Rec/rec.games.roguelike.development/2010-09/msg00050.html */
+# define _X_OPEN_SOURCE_EXTENDED
+
+# if HAVE_NCURSESW_NCURSES_H
+#  include <ncursesw/ncurses.h>
+# elif HAVE_NCURSES_H
+#  include <ncurses.h>
+# endif
 
 # define OPING_GREEN 1
 # define OPING_YELLOW 2
 # define OPING_RED 3
+# define OPING_GREEN_HIST 4
+# define OPING_YELLOW_HIST 5
+# define OPING_RED_HIST 6
+
+static char const * const hist_symbols_utf8[] = {
+	"▁", "▂", "▃", "▄", "▅", "▆", "▇", "█" };
+static size_t const hist_symbols_utf8_num = sizeof (hist_symbols_utf8)
+	/ sizeof (hist_symbols_utf8[0]);
+
+/* scancodes for 6 levels of horizontal bars, ncurses-specific */
+/* those are not the usual constants because those are not constant */
+static int const hist_symbols_acs[] = {
+	115, /* ACS_S9 "⎽" */
+	114, /* ACS_S7 "⎼" */
+	113, /* ACS_S5 "─" */
+	112, /* ACS_S3 "⎻" */
+	111  /* ACS_S1 "⎺" */
+};
+static size_t const hist_symbols_acs_num = sizeof (hist_symbols_acs)
+	/ sizeof (hist_symbols_acs[0]);
+
+/* use different colors without a background for scancodes */
+static int const hist_colors_utf8[] = {
+	OPING_GREEN_HIST, OPING_YELLOW_HIST, OPING_RED_HIST };
+static int const hist_colors_acs[] = {
+	OPING_GREEN, OPING_YELLOW, OPING_RED };
+/* assuming that both arrays are the same size */
+static size_t const hist_colors_num = sizeof (hist_colors_utf8)
+	/ sizeof (hist_colors_utf8[0]);
 #endif
 
 #include "oping.h"
@@ -89,6 +128,10 @@
 # define _POSIX_SAVED_IDS 0
 #endif
 
+#ifndef IPTOS_MINCOST
+# define IPTOS_MINCOST 0x02
+#endif
+
 /* Remove GNU specific __attribute__ settings when using another compiler */
 #if !__GNUC__
 # define __attribute__(x) /**/
@@ -121,6 +164,10 @@ static char   *opt_filename   = NULL;
 static int     opt_count      = -1;
 static int     opt_send_ttl   = 64;
 static uint8_t opt_send_qos   = 0;
+static double  opt_exit_status_threshold = 1.0;
+#if USE_NCURSES
+static int     opt_utf8       = 0;
+#endif
 
 static int host_num = 0;
 
@@ -265,6 +312,11 @@ static void usage_exit (const char *name, int status) /* {{{ */
 			"  -I srcaddr   source address\n"
 			"  -D device    outgoing interface name\n"
 			"  -f filename  filename to read hosts from\n"
+#if USE_NCURSES
+			"  -u / -U      force / disable UTF-8 output\n"
+#endif
+			"  -Z percent   Exit with non-zero exit status if more than this percentage of\n"
+			"               probes timed out. (default: never)\n"
 
 			"\noping "PACKAGE_VERSION", http://verplant.org/liboping/\n"
 			"by Florian octo Forster <octo@verplant.org>\n"
@@ -467,7 +519,11 @@ static int read_options (int argc, char **argv) /* {{{ */
 
 	while (1)
 	{
-		optchar = getopt (argc, argv, "46c:hi:I:t:Q:f:D:");
+		optchar = getopt (argc, argv, "46c:hi:I:t:Q:f:D:Z:"
+#if USE_NCURSES
+				"uU"
+#endif
+				);
 
 		if (optchar == -1)
 			break;
@@ -538,6 +594,33 @@ static int read_options (int argc, char **argv) /* {{{ */
 				set_opt_send_qos (optarg);
 				break;
 
+#if USE_NCURSES
+			case 'u':
+				opt_utf8 = 2;
+				break;
+			case 'U':
+				opt_utf8 = 1;
+				break;
+#endif
+
+			case 'Z':
+			{
+				char *endptr = NULL;
+				double tmp;
+
+				errno = 0;
+				tmp = strtod (optarg, &endptr);
+				if ((errno != 0) || (endptr == NULL) || (*endptr != 0) || (tmp < 0.0) || (tmp > 100.0))
+				{
+					fprintf (stderr, "Ignoring invalid -Z argument: %s\n", optarg);
+					fprintf (stderr, "The \"-Z\" option requires a numeric argument between 0 and 100.\n");
+				}
+				else
+					opt_exit_status_threshold = tmp / 100.0;
+
+				break;
+			}
+
 			case 'h':
 				usage_exit (argv[0], 0);
 				break;
@@ -597,12 +680,120 @@ static void time_calc (struct timespec *ts_dest, /* {{{ */
 } /* }}} void time_calc */
 
 #if USE_NCURSES
-static int update_stats_from_context (ping_context_t *ctx) /* {{{ */
+static _Bool has_utf8() /* {{{ */
 {
+# if HAVE_NCURSESW_NCURSES_H
+	if (!opt_utf8)
+	{
+		/* Automatically determine */
+		if (strcasecmp ("UTF-8", nl_langinfo (CODESET)) == 0)
+			opt_utf8 = 2;
+		else
+			opt_utf8 = 1;
+	}
+	return ((_Bool) (opt_utf8 - 1));
+# else
+	return (0);
+# endif
+} /* }}} _Bool has_utf8 */
+
+static int update_prettyping_graph (ping_context_t *ctx, /* {{{ */
+		double latency, unsigned int sequence)
+{
+	int color = OPING_RED;
+	char const *symbol = "!";
+	int symbolc = '!';
+
+	int x_max;
+	int x_pos;
+
+	x_max = getmaxx (ctx->window);
+	x_pos = ((sequence - 1) % (x_max - 4)) + 2;
+
+	if (latency >= 0.0)
+	{
+		double ratio;
+
+		size_t symbols_num = hist_symbols_acs_num;
+		size_t colors_num = 1;
+
+		size_t index_symbols;
+		size_t index_colors;
+		size_t intensity;
+
+		/* latency is in milliseconds, opt_interval is in seconds. */
+		ratio = (latency * 0.001) / opt_interval;
+		if (ratio > 1) {
+			ratio = 1.0;
+		}
+
+		if (has_utf8 ())
+			symbols_num = hist_symbols_utf8_num;
+
+		if (has_colors () == TRUE)
+			colors_num = hist_colors_num;
+
+		intensity = (size_t) (ratio * ((double) (symbols_num * colors_num)));
+		if (intensity >= (symbols_num * colors_num))
+			intensity = (symbols_num * colors_num) - 1;
+
+		index_symbols = intensity % symbols_num;
+		assert (index_symbols < symbols_num);
+
+		index_colors = intensity / symbols_num;
+		assert (index_colors < colors_num);
+
+		if (has_utf8())
+		{
+			color = hist_colors_utf8[index_colors];
+			symbol = hist_symbols_utf8[index_symbols];
+		}
+		else
+		{
+			color = hist_colors_acs[index_colors];
+			symbolc = hist_symbols_acs[index_symbols] | A_ALTCHARSET;
+		}
+	}
+	else /* if (!(latency >= 0.0)) */
+		wattron (ctx->window, A_BOLD);
+
+	if (has_colors () == TRUE)
+		wattron (ctx->window, COLOR_PAIR(color));
+
+	if (has_utf8())
+		mvwprintw (ctx->window, /* y = */ 3, /* x = */ x_pos, symbol);
+	else
+		mvwaddch (ctx->window, /* y = */ 3, /* x = */ x_pos, symbolc);
+
+	if (has_colors () == TRUE)
+		wattroff (ctx->window, COLOR_PAIR(color));
+
+	/* Use negation here to handle NaN correctly. */
+	if (!(latency >= 0.0))
+		wattroff (ctx->window, A_BOLD);
+
+	wprintw (ctx->window, " ");
+	return (0);
+} /* }}} int update_prettyping_graph */
+
+static int update_stats_from_context (ping_context_t *ctx, pingobj_iter_t *iter) /* {{{ */
+{
+	double latency = -1.0;
+	size_t buffer_len = sizeof (latency);
+
+	ping_iterator_get_info (iter, PING_INFO_LATENCY,
+			&latency, &buffer_len);
+
+	unsigned int sequence = 0;
+	buffer_len = sizeof (sequence);
+	ping_iterator_get_info (iter, PING_INFO_SEQUENCE,
+			&sequence, &buffer_len);
+
+
 	if ((ctx == NULL) || (ctx->window == NULL))
 		return (EINVAL);
 
-	werase (ctx->window);
+	/* werase (ctx->window); */
 
 	box (ctx->window, 0, 0);
 	wattron (ctx->window, A_BOLD);
@@ -623,7 +814,7 @@ static int update_stats_from_context (ping_context_t *ctx) /* {{{ */
 
 		average = context_get_average (ctx);
 		deviation = context_get_stddev (ctx);
-			
+
 		mvwprintw (ctx->window, /* y = */ 2, /* x = */ 2,
 				"rtt min/avg/max/sdev = %.3f/%.3f/%.3f/%.3f ms",
 				ctx->latency_min,
@@ -632,6 +823,8 @@ static int update_stats_from_context (ping_context_t *ctx) /* {{{ */
 				deviation);
 	}
 
+	update_prettyping_graph (ctx, latency, sequence);
+
 	wrefresh (ctx->window);
 
 	return (0);
@@ -648,7 +841,7 @@ static int on_resize (pingobj_t *ping) /* {{{ */
 	if ((height < 1) || (width < 1))
 		return (EINVAL);
 
-	main_win_height = height - (4 * host_num);
+	main_win_height = height - (5 * host_num);
 	wresize (main_win, main_win_height, /* width = */ width);
 	/* Allow scrolling */
 	scrollok (main_win, TRUE);
@@ -672,9 +865,9 @@ static int on_resize (pingobj_t *ping) /* {{{ */
 			delwin (context->window);
 			context->window = NULL;
 		}
-		context->window = newwin (/* height = */ 4,
+		context->window = newwin (/* height = */ 5,
 				/* width = */ width,
-				/* y = */ main_win_height + (4 * context->index),
+				/* y = */ main_win_height + (5 * context->index),
 				/* x = */ 0);
 	}
 
@@ -722,9 +915,12 @@ static int pre_loop_hook (pingobj_t *ping) /* {{{ */
 		init_pair (OPING_GREEN,  COLOR_GREEN,  /* default = */ 0);
 		init_pair (OPING_YELLOW, COLOR_YELLOW, /* default = */ 0);
 		init_pair (OPING_RED,    COLOR_RED,    /* default = */ 0);
+		init_pair (OPING_GREEN_HIST,  COLOR_GREEN,  COLOR_BLACK);
+		init_pair (OPING_YELLOW_HIST, COLOR_YELLOW, COLOR_GREEN);
+		init_pair (OPING_RED_HIST,    COLOR_RED,    COLOR_YELLOW);
 	}
 
-	main_win_height = height - (4 * host_num);
+	main_win_height = height - (5 * host_num);
 	main_win = newwin (/* height = */ main_win_height,
 			/* width = */ width,
 			/* y = */ 0, /* x = */ 0);
@@ -751,9 +947,9 @@ static int pre_loop_hook (pingobj_t *ping) /* {{{ */
 			delwin (context->window);
 			context->window = NULL;
 		}
-		context->window = newwin (/* height = */ 4,
+		context->window = newwin (/* height = */ 5,
 				/* width = */ width,
-				/* y = */ main_win_height + (4 * context->index),
+				/* y = */ main_win_height + (5 * context->index),
 				/* x = */ 0);
 	}
 
@@ -946,14 +1142,18 @@ static void update_host_hook (pingobj_iter_t *iter, /* {{{ */
 	}
 
 #if USE_NCURSES
-	update_stats_from_context (context);
+	update_stats_from_context (context, iter);
 	wrefresh (main_win);
 #endif
 } /* }}} void update_host_hook */
 
+/* Prints statistics for each host, cleans up the contexts and returns the
+ * number of hosts which failed to return more than the fraction
+ * opt_exit_status_threshold of pings. */
 static int post_loop_hook (pingobj_t *ping) /* {{{ */
 {
 	pingobj_iter_t *iter;
+	int failure_count = 0;
 
 #if USE_NCURSES
 	endwin ();
@@ -973,6 +1173,13 @@ static int post_loop_hook (pingobj_t *ping) /* {{{ */
 				context_get_packet_loss (context),
 				context->latency_total);
 
+		{
+			double pct_failed = 1.0 - (((double) context->req_rcvd)
+					/ ((double) context->req_sent));
+			if (pct_failed > opt_exit_status_threshold)
+				failure_count++;
+		}
+
 		if (context->req_rcvd != 0)
 		{
 			double average;
@@ -992,7 +1199,7 @@ static int post_loop_hook (pingobj_t *ping) /* {{{ */
 		context_destroy (context);
 	}
 
-	return (0);
+	return (failure_count);
 } /* }}} int post_loop_hook */
 
 int main (int argc, char **argv) /* {{{ */
@@ -1026,6 +1233,7 @@ int main (int argc, char **argv) /* {{{ */
 	}
 #endif
 
+        setlocale(LC_ALL, "");
 	optind = read_options (argc, argv);
 
 #if !_POSIX_SAVED_IDS
@@ -1230,7 +1438,12 @@ int main (int argc, char **argv) /* {{{ */
 			return (1);
 		}
 
-		if (ping_send (ping) < 0)
+		status = ping_send (ping);
+		if (status == -EINTR)
+		{
+			continue;
+		}
+		else if (status < 0)
 		{
 			fprintf (stderr, "ping_send failed: %s\n",
 					ping_get_error (ping));
@@ -1263,16 +1476,15 @@ int main (int argc, char **argv) /* {{{ */
 		/* printf ("Sleeping for %i.%09li seconds\n", (int) ts_wait.tv_sec, ts_wait.tv_nsec); */
 		while ((status = nanosleep (&ts_wait, &ts_wait)) != 0)
 		{
-			if (errno != EINTR)
+			if (errno == EINTR)
+			{
+				continue;
+			}
+			else
 			{
 				perror ("nanosleep");
 				break;
 			}
-			else if (opt_count == 0)
-			{
-				/* sigint */
-				break;
-			}
 		}
 
 		post_sleep_hook (ping);
@@ -1281,11 +1493,19 @@ int main (int argc, char **argv) /* {{{ */
 			opt_count--;
 	} /* while (opt_count != 0) */
 
-	post_loop_hook (ping);
+	/* Returns the number of failed hosts according to -Z. */
+	status = post_loop_hook (ping);
 
 	ping_destroy (ping);
 
-	return (0);
+	if (status == 0)
+		exit (EXIT_SUCCESS);
+	else
+	{
+		if (status > 255)
+			status = 255;
+		exit (status);
+	}
 } /* }}} int main */
 
 /* vim: set fdm=marker : */
